# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Email Framework
"""

from __future__ import unicode_literals, absolute_import

import os
import smtplib
import logging
from email.charset import Charset
from email.message import Message
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText

import six

from mako.template import Template
from mako.lookup import TemplateLookup
from mako.exceptions import TopLevelLookupException

from rattail import exceptions
from rattail.core import UNSPECIFIED
from rattail.files import resource_path
from rattail.util import import_module_path
from rattail.time import localtime, make_utc


# NOTE: this bit of magic was stolen from Django
# Don't BASE64-encode UTF-8 messages so that we avoid unwanted attention from
# some spam filters.
utf8_charset = Charset('utf-8')
utf8_charset.body_encoding = None  # Python defaults to BASE64

log = logging.getLogger(__name__)


def send_email(config, key, data={}, attachments=[],
               fallback_key=None, default_subject=None,
               enabled=UNSPECIFIED, **kwargs):
    """
    Send an email message of the given type, per config, with the given data
    and/or attachments.
    """
    # TODO: should let config override which handler we use
    handler = EmailHandler(config)
    email = handler.get_email(key, fallback_key=fallback_key,
                              default_subject=default_subject)

    if enabled is UNSPECIFIED:
        enabled = handler.get_enabled(email)

    if enabled:
        kwargs['attachments'] = attachments
        handler.send_message(email, data, **kwargs)
    else:
        log.debug("skipping email of type '%s' per config", key)


# TODO: deprecate / remove this (used only for tailbone preview?)
def deliver_message(config, key, msg, recipients=UNSPECIFIED):
    """
    Deliver an email message using the given SMTP configuration.
    """
    if recipients is UNSPECIFIED:
        recips = set()
        to = msg.get_all('To')
        if to:
            recips = recips.union(set(to))
        cc = msg.get_all('Cc')
        if cc:
            recips = recips.union(set(cc))
        bcc = msg.get_all('Bcc')
        if bcc:
            recips = recips.union(set(bcc))
    else:
        recips = set(recipients)
    if not recips:
        raise RuntimeError("No recipients for email: {0}".format(repr(msg)))

    server = config.get('rattail.mail', 'smtp.server', default='localhost')
    username = config.get('rattail.mail', 'smtp.username')
    password = config.get('rattail.mail', 'smtp.password')

    if config.getbool('rattail.mail', 'send_feedback_only', usedb=False, default=False):
        send = key == 'user_feedback'
    else:
        send = config.getbool('rattail.mail', 'send_emails', usedb=False, default=True)

    if send:

        log.debug("attempting to send mail of type: %s", key)
        log.debug("connecting to server: %s", server)
        session = smtplib.SMTP(server)
        if username and password:
            result = session.login(username, password)
            log.debug("login() result is: %s", repr(result))

        result = session.sendmail(msg['From'], recips, msg.as_string())
        log.debug("sendmail() result is: %s", repr(result))
        session.quit()
        return True

    log.debug("config says no emails for '%s', but would have sent one to: %s", key, recips)
    return False


class EmailHandler(object):
    """
    Base class and default implementation for email handlers.
    """

    def __init__(self, config):
        self.config = config
        self.enum = self.config.get_enum()

    def get_email(self, key, fallback_key=None, **kwargs):
        """
        Return an email instance of the given type.
        """
        for email in self.iter_emails():
            if email.key == key or email.__name__ == key:
                return email(self.config, key, fallback_key, **kwargs)
        return Email(self.config, key, fallback_key, **kwargs)

    def iter_emails(self):
        """
        Iterate over all available email types.
        """
        for module in self.config.getlist('rattail.mail', 'emails', default=['rattail.emails']):
            module = import_module_path(module)
            for name in dir(module):
                obj = getattr(module, name)
                if (isinstance(obj, type) and issubclass(obj, Email)
                    and not obj.abstract and obj is not Email):
                    yield obj

    def get_enabled(self, email):
        return email.get_enabled()

    def send_message(self, email, data, **kwargs):
        msg = self.make_message(email, data, **kwargs)

        if self.config.getbool('rattail.mail', 'record_attempts', default=False):
            attempt = self.record_attempt(email, msg)
            try:
                self.deliver_message(email, msg)
            except Exception as e:
                self.record_failure(attempt, e)
            else:
                self.record_success(attempt)

        else: # don't record attempts
            self.deliver_message(email, msg)

    def record_attempt(self, email, msg):
        from rattail.db import Session, model

        session = Session()

        attempt = model.EmailAttempt()
        attempt.key = email.key
        attempt.sender = msg['From']
        attempt.to = msg.get_all('To')
        attempt.cc = msg.get_all('Cc')
        attempt.bcc = msg.get_all('Bcc')
        attempt.subject = msg['Subject']
        attempt.sent = make_utc()
        attempt.status_code = self.enum.EMAIL_ATTEMPT_CREATED

        session.add(attempt)
        session.commit()
        session.close()
        # session.expunge(attempt)
        return attempt

    def record_failure(self, attempt, error):
        from rattail.db import Session

        session = Session()
        attempt = session.merge(attempt)
        attempt.status_code = self.enum.EMAIL_ATTEMPT_FAILURE
        attempt.status_text = six.text_type(error)
        session.commit()
        session.close()

    def record_success(self, attempt):
        from rattail.db import Session

        session = Session()
        attempt = session.merge(attempt)
        attempt.status_code = self.enum.EMAIL_ATTEMPT_SUCCESS
        session.commit()
        session.close()

    def make_message(self, email, data, **kwargs):
        context = self.make_context(**data)
        return email.make_message(context, **kwargs)

    def make_context(self, **context):
        context['rattail_config'] = self.config
        context['app_title'] = self.config.app_title(default="Rattail")
        context['localtime'] = localtime
        return context

    def deliver_message(self, email, msg, recipients=UNSPECIFIED):
        """
        Deliver an email message using the given SMTP configuration.
        """
        if recipients is UNSPECIFIED:
            recips = set()
            to = msg.get_all('To')
            if to:
                recips = recips.union(set(to))
            cc = msg.get_all('Cc')
            if cc:
                recips = recips.union(set(cc))
            bcc = msg.get_all('Bcc')
            if bcc:
                recips = recips.union(set(bcc))
        else:
            recips = set(recipients)
        if not recips:
            raise RuntimeError("No recipients for email: {0}".format(repr(msg)))

        server = self.config.get('rattail.mail', 'smtp.server', default='localhost')
        username = self.config.get('rattail.mail', 'smtp.username')
        password = self.config.get('rattail.mail', 'smtp.password')

        if self.config.getbool('rattail.mail', 'send_feedback_only', usedb=False, default=False):
            send = email.key == 'user_feedback'
        else:
            send = self.config.getbool('rattail.mail', 'send_emails', usedb=False, default=True)

        if send:

            log.debug("attempting to send mail of type: %s", email.key)
            log.debug("connecting to server: %s", server)
            session = smtplib.SMTP(server)
            if username and password:
                result = session.login(username, password)
                log.debug("login() result is: %s", repr(result))

            result = session.sendmail(msg['From'], recips, msg.as_string())
            log.debug("sendmail() result is: %s", repr(result))
            session.quit()
            return True

        log.debug("config says no emails for '%s', but would have sent one to: %s", email.key, recips)
        return False


class Email(object):
    # Note: The docstring of an email is leveraged by code, hence this odd one.
    """
    (This email has no description.)
    """
    key = None
    fallback_key = None
    abstract = False
    default_prefix = "[rattail]"
    default_subject = "Automated message"

    # Whether or not the email's :attr:`to` attribute is dynamically determined
    # at run-time, i.e. via some logic other than typical reading from config.
    dynamic_to = False
    dynamic_to_help = None

    def __init__(self, config, key=None, fallback_key=None, default_subject=None):
        self.config = config
        self.enum = config.get_enum()

        if key:
            self.key = key
        elif not self.key:
            self.key = self.__class__.__name__
            if self.key == 'Email':
                raise exceptions.ConfigurationError("Email instance has no key: {0}".format(repr(self)))

        if fallback_key:
            self.fallback_key = fallback_key
        if default_subject:
            self.default_subject = default_subject

        templates = config.getlist('rattail.mail', 'templates')
        if templates:
            templates = [resource_path(p) for p in templates]
        self.templates = TemplateLookup(directories=templates)

    def obtain_sample_data(self, request):
        """
        This method is responsible for obtaining the full set of sample data,
        to be used as context when generating a preview for the email.

        Note, you normally should not override this method!  Please see also
        the :meth:`sample_data()` method.
        """
        return self.sample_data(request)

    def sample_data(self, request):
        """
        This method can return a dict of sample data, to be used as context
        when generating a preview for the email.  Subclasses are welcome to
        override this method.
        """
        return {}

    def get_enabled(self):
        """
        Get the enabled flag for the email's message type.
        """
        enabled = self.config.getbool('rattail.mail', '{0}.enabled'.format(self.key))
        if enabled is not None:
            return enabled
        enabled = self.config.getbool('rattail.mail', 'default.enabled')
        if enabled is not None:
            return enabled
        return self.config.getbool('rattail.mail', 'send_emails', default=True)

    def get_sender(self):
        """
        Returns the value for the message's ``From:`` header.

        :rtype: str
        """
        sender = self.config.get('rattail.mail', '{0}.from'.format(self.key))
        if not sender:
            sender = self.config.get('rattail.mail', 'default.from')
            if not sender:
                raise exceptions.SenderNotFound(self.key)
        return sender

    def get_replyto(self):
        """
        Get the Reply-To address for the message.
        """
        replyto = self.config.get('rattail.mail', '{0}.replyto'.format(self.key))
        if not replyto:
            replyto = self.config.get('rattail.mail', 'default.replyto')
        return replyto

    def get_recips(self, type_='to'):
        """
        Returns a list of recipients of the given type for the message.

        :param type_: Must be one of: ``('to', 'cc', 'bcc')``.

        :rtype: list
        """
        try:
            if type_.lower() not in ('to', 'cc', 'bcc'):
                raise Exception
        except:
            raise ValueError("Recipient type must be one of ('to', 'cc', 'bcc'); "
                             "not: {0}".format(repr(type_)))
        type_ = type_.lower()
        recips = self.config.getlist('rattail.mail', '{0}.{1}'.format(self.key, type_))
        if not recips:
            recips = self.config.getlist('rattail.mail', 'default.{0}'.format(type_))
        return recips

    def get_prefix(self, data={}, magic=True):
        """
        Returns a string to be used as the subject prefix for the message.

        :rtype: str
        """
        prefix = self.config.get('rattail.mail', '{0}.prefix'.format(self.key))
        if not prefix:
            prefix = self.config.get('rattail.mail', 'default.prefix')
        prefix = prefix or self.default_prefix
        if magic and not self.config.production():
            prefix = "[STAGE] {}".format(prefix)
        return prefix

    def get_subject(self, data={}, render=True, template=UNSPECIFIED):
        """
        Returns the base value for the message's subject header, i.e. minus
        prefix.

        :rtype: str
        """
        if template is UNSPECIFIED:
            template = self.config.get('rattail.mail', '{}.subject'.format(self.key),
                                       default=self.default_subject)
            if not template:
                template = self.config.get('rattail.mail', 'default.subject')
        if template and render:
            return Template(template).render(**data)
        return template

    def get_complete_subject(self, data={}, render=True, prefix=UNSPECIFIED, template=UNSPECIFIED):
        """
        Returns the value for the message's ``Subject:`` header, i.e. the base
        subject with the prefix applied.  Note that config may provide the
        complete subject also, in which case the prefix and base subject are
        not considered.

        :rtype: str
        """
        if prefix is UNSPECIFIED:
            prefix = self.get_prefix(data)
        prefix = (prefix or "").rstrip()
        if prefix:
            prefix = "{} ".format(prefix)
        return "{}{}".format(prefix, self.get_subject(data, render=render, template=template))

    def get_template(self, type_):
        """
        Locate and return the Mako email template of the given type
        (e.g. 'html'), or ``None`` if no such template can be found.
        """
        try:
            return self.templates.get_template('{0}.{1}.mako'.format(self.key, type_))
        except TopLevelLookupException:
            if self.fallback_key:
                try:
                    return self.templates.get_template('{0}.{1}.mako'.format(self.fallback_key, type_))
                except TopLevelLookupException:
                    pass

    def normalize_attachments(self, attachments):
        normalized = []
        for attachment in attachments:
            if isinstance(attachment, six.string_types):
                attachment = self.normalize_attachment(attachment)
            normalized.append(attachment)
        return normalized

    ATTACHMENT_MIME_MAP = {
        '.doc': 'application/msword',
        '.pdf': 'pdf',
        '.xls': 'vnd.ms-excel',
    }

    def normalize_attachment(self, path):
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext == '.csv':
            with open(path, 'rb') as f:
                part = MIMEText(f.read(), 'csv', 'utf_8')
            filename = os.path.basename(path)
            part.add_header('Content-Disposition', 'attachment; filename="{}"'.format(filename))
            return part
        else:
            mimetype = self.ATTACHMENT_MIME_MAP.get(ext)
            if mimetype:
                with open(path, 'rb') as f:
                    part = MIMEApplication(f.read(), mimetype)
                filename = os.path.basename(path)
                part.add_header('Content-Disposition', 'attachment; filename="{}"'.format(filename))
                return part
        raise ValueError("Magic is not (yet) supported, please construct your own attachments for file: {}".format(path))

    def make_message(self, data={}, attachments=[], inlines=[],
                     subject_prefix=UNSPECIFIED, subject_template=UNSPECIFIED,
                     sender=UNSPECIFIED, replyto=UNSPECIFIED,
                     to=UNSPECIFIED, cc=UNSPECIFIED, bcc=UNSPECIFIED):
        """
        Returns a proper email ``Message`` instance which may be sent via SMTP.
        """
        txt_template = self.get_template('txt')
        html_template = self.get_template('html')
        attachments = self.normalize_attachments(attachments)

        # TODO: provide more defaults?
        data.setdefault('six', six)

        if txt_template and html_template:

            txt_part = MIMEText(txt_template.render(**data), _charset='utf_8')

            html_part = MIMEText(html_template.render(**data), _subtype='html', _charset='utf_8')
            if inlines:
                html_part = MIMEMultipart(_subtype='related', _subparts=[html_part] + inlines)

            msg = MIMEMultipart(_subtype='alternative', _subparts=[txt_part, html_part])
            if attachments:
                msg = MIMEMultipart(_subtype='mixed', _subparts=[msg] + attachments)

        elif txt_template:

            msg = MIMEText(txt_template.render(**data), _charset='utf_8')
            if attachments:
                msg = MIMEMultipart(_subtype='mixed', _subparts=[msg] + attachments)

        elif html_template:

            msg = SafeMIMEText(html_template.render(**data), 'html', utf8_charset)
            if inlines:
                msg = MIMEMultipart(_subtype='related', _subparts=[msg] + inlines)
            if attachments:
                msg = MIMEMultipart(_subtype='mixed', _subparts=[msg] + attachments)

        else:
            raise exceptions.MailTemplateNotFound(self.key)

        self.add_headers(msg, data=data,
                         subject_prefix=subject_prefix, subject_template=subject_template,
                         sender=sender, replyto=replyto, to=to, cc=cc, bcc=bcc)
        return msg

    def add_headers(self, msg, data={},
                    subject_prefix=UNSPECIFIED, subject_template=UNSPECIFIED,
                    sender=UNSPECIFIED, replyto=UNSPECIFIED,
                    to=UNSPECIFIED, cc=UNSPECIFIED, bcc=UNSPECIFIED):
        """
        Adds headers for to/from addresses etc. to message
        """
        # subject/from
        msg['Subject'] = self.get_complete_subject(data, prefix=subject_prefix,
                                                   template=subject_template)
        if sender is UNSPECIFIED:
            sender = self.get_sender()
        msg['From'] = sender

        # reply-to
        if replyto is UNSPECIFIED:
            replyto = self.get_replyto()
        if replyto:
            msg.add_header('Reply-To', replyto)

        # recipients
        force_to = self.config.getlist('rattail.mail', 'force_to', usedb=False)
        if force_to:
            to = force_to
            cc = None
            bcc = None
        else:
            if to is UNSPECIFIED:
                to = self.get_recips('to')
            if cc is UNSPECIFIED:
                cc = self.get_recips('cc')
            if bcc is UNSPECIFIED:
                bcc = self.get_recips('bcc')
        if not (to or cc or bcc):
            raise exceptions.RecipientsNotFound(self.key)
        if to:
            for recip in to:
                msg['To'] = recip
        if cc:
            for recip in cc:
                msg['Cc'] = recip
        if bcc:
            for recip in bcc:
                msg['Bcc'] = recip


# NOTE: this bit of magic was stolen from Django
class SafeMIMEText(MIMEText):

    def __init__(self, text, subtype, charset):
        self.encoding = charset
        if charset == 'utf-8':
            # Unfortunately, Python doesn't support setting a Charset instance
            # as MIMEText init parameter (http://bugs.python.org/issue16324).
            # We do it manually and trigger re-encoding of the payload.
            MIMEText.__init__(self, text, subtype, None)
            del self['Content-Transfer-Encoding']
            # TODO: i don't personally need this yet, if ever?
            # # Workaround for versions without http://bugs.python.org/issue19063
            # if (3, 2) < sys.version_info < (3, 3, 4):
            #     payload = text.encode(utf8_charset.output_charset)
            #     self._payload = payload.decode('ascii', 'surrogateescape')
            #     self.set_charset(utf8_charset)
            # else:
            #     self.set_payload(text, utf8_charset)
            self.set_payload(text, utf8_charset)
            self.replace_header('Content-Type', 'text/%s; charset="%s"' % (subtype, charset))
        else:
            MIMEText.__init__(self, text, subtype, charset)



# -*- coding: utf-8 -*-
"""add tempmon

Revision ID: 9740f0dcf84b
Revises: 3f266c89a3d4
Create Date: 2016-11-21 20:26:08.789247

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '9740f0dcf84b'
down_revision = u'3f266c89a3d4'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # tempmon_client
    op.create_table('tempmon_client',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('config_key', sa.String(length=50), nullable=False),
                    sa.Column('hostname', sa.String(length=255), nullable=False),
                    sa.Column('location', sa.String(length=255), nullable=True),
                    sa.Column('enabled', sa.Boolean(), nullable=False),
                    sa.Column('online', sa.Boolean(), nullable=False),
                    sa.UniqueConstraint('config_key', name='tempmon_client_uq_config_key'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # tempmon_probe
    op.create_table('tempmon_probe',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('client_uuid', sa.String(length=32), nullable=False),
                    sa.Column('config_key', sa.String(length=50), nullable=False),
                    sa.Column('appliance_type', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=False),
                    sa.Column('device_path', sa.String(length=255), nullable=True),
                    sa.Column('good_temp_min', sa.Integer(), nullable=False),
                    sa.Column('good_temp_max', sa.Integer(), nullable=False),
                    sa.Column('critical_temp_min', sa.Integer(), nullable=False),
                    sa.Column('critical_temp_max', sa.Integer(), nullable=False),
                    sa.Column('status_alert_timeout', sa.Integer(), nullable=False),
                    sa.Column('therm_status_timeout', sa.Integer(), nullable=False),
                    sa.Column('enabled', sa.Boolean(), nullable=False),
                    sa.Column('status', sa.Integer(), nullable=True),
                    sa.Column('status_changed', sa.DateTime(), nullable=True),
                    sa.Column('status_alert_sent', sa.DateTime(), nullable=True),
                    sa.ForeignKeyConstraint(['client_uuid'], [u'tempmon_client.uuid'], name=u'tempmon_probe_fk_client'),
                    sa.UniqueConstraint('config_key', name='tempmon_probe_uq_config_key'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # tempmon_reading
    op.create_table('tempmon_reading',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('client_uuid', sa.String(length=32), nullable=False),
                    sa.Column('probe_uuid', sa.String(length=32), nullable=False),
                    sa.Column('taken', sa.DateTime(), nullable=False),
                    sa.Column('degrees_f', sa.Numeric(precision=7, scale=4), nullable=False),
                    sa.ForeignKeyConstraint(['client_uuid'], [u'tempmon_client.uuid'], name=u'tempmon_reading_fk_client'),
                    sa.ForeignKeyConstraint(['probe_uuid'], [u'tempmon_probe.uuid'], name=u'tempmon_reading_fk_probe'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # tempmon
    op.drop_table('tempmon_reading')
    op.drop_table('tempmon_probe')
    op.drop_table('tempmon_client')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Trainwreck model importers
"""

from __future__ import unicode_literals, absolute_import

from rattail import importing
from rattail.time import make_utc
from .util import ToOrFromTrainwreck


class ToTrainwreck(ToOrFromTrainwreck, importing.ToSQLAlchemy):
    """
    Base class for all Trainwreck model importers
    """
    key = 'uuid'


class TransactionImporter(ToTrainwreck):
    """
    Transaction data importer
    """
    # NOTE: you must subclass this and define the model
    # model_class = trainwreck.Transaction

    def cache_query(self):
        query = super(TransactionImporter, self).cache_query()
        return query.filter(self.model_class.end_time >= make_utc(self.start_time))\
                    .filter(self.model_class.end_time < make_utc(self.end_time))


class TransactionItemImporter(ToTrainwreck):
    """
    Transaction item data importer
    """
    # NOTE: you must subclass this and define the model
    # model_class = trainwreck.TransactionItem

    # NOTE: subclass must also define this
    # transaction_class = trainwreck.Transaction

    def cache_query(self):
        query = super(TransactionItemImporter, self).cache_query()
        return query.join(self.transaction_class)\
                    .filter(self.transaction_class.end_time >= make_utc(self.start_time))\
                    .filter(self.transaction_class.end_time < make_utc(self.end_time))



# -*- coding: utf-8; -*-
"""add batch.params

Revision ID: 5ff1dddc0a30
Revises: 4f026ae58751
Create Date: 2019-10-08 17:17:23.616763

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '5ff1dddc0a30'
down_revision = '4f026ae58751'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch*
    op.add_column('batch_handheld', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('batch_importer', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('batch_inventory', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('batch_newproduct', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('batch_pricing', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('batch_product', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('label_batch', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('purchase_batch', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('vendor_catalog', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))
    op.add_column('vendor_invoice', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))


def downgrade():

    # batch*
    op.drop_column('vendor_invoice', 'params')
    op.drop_column('vendor_catalog', 'params')
    op.drop_column('purchase_batch', 'params')
    op.drop_column('label_batch', 'params')
    op.drop_column('batch_product', 'params')
    op.drop_column('batch_pricing', 'params')
    op.drop_column('batch_newproduct', 'params')
    op.drop_column('batch_inventory', 'params')
    op.drop_column('batch_importer', 'params')
    op.drop_column('batch_handheld', 'params')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Models for vendor invoice batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import Base, Vendor, FileBatchMixin, ProductBatchRowMixin


class VendorInvoice(FileBatchMixin, Base):
    """
    Vendor invoice, the source data file of which has been provided by a user,
    and may be further processed in some site-specific way.
    """
    __tablename__ = 'vendor_invoice'
    __batchrow_class__ = 'VendorInvoiceRow'

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'],
                                    name='vendor_invoice_fk_vendor'),
            )

    parser_key = sa.Column(sa.String(length=100), nullable=False, doc="""
    The key of the parser used to parse the contents of the data file.
    """)

    vendor_uuid = sa.Column(sa.String(length=32), nullable=False)

    vendor = relationship(Vendor, doc="""
    Reference to the :class:`Vendor` to which the invoice pertains.
    """)

    invoice_date = sa.Column(sa.Date(), nullable=True, doc="""
    Invoice date, as determined by the invoice data file.
    """)

    purchase_order_number = sa.Column(sa.String(length=20), nullable=True, doc="""
    Purchase order number, e.g. for cross-reference with another system.  Custom
    batch handlers may populate and leverage this field, but the default handler
    does not.
    """)


class VendorInvoiceRow(ProductBatchRowMixin, Base):
    """
    Row of data within a vendor invoice.
    """
    __tablename__ = 'vendor_invoice_row'
    __batch_class__ = VendorInvoice

    STATUS_OK = 1
    STATUS_NOT_IN_DB = 2
    STATUS_NOT_IN_PURCHASE = 3
    STATUS_NOT_IN_INVOICE = 4
    STATUS_DIFFERS_FROM_PURCHASE = 5
    STATUS_COST_NOT_IN_DB = 6
    STATUS_NO_CASE_QUANTITY = 7
    STATUS_UNIT_COST_DIFFERS = 8

    STATUS = {
        STATUS_OK:                      "ok",
        STATUS_NOT_IN_DB:               "product not found",
        STATUS_COST_NOT_IN_DB:          "product found but not cost",
        STATUS_NOT_IN_PURCHASE:         "present only in invoice",
        STATUS_NOT_IN_INVOICE:          "present only in PO",
        STATUS_DIFFERS_FROM_PURCHASE:   "invoice and PO differ",
        STATUS_NO_CASE_QUANTITY:        "case quantity not known",
        STATUS_UNIT_COST_DIFFERS:       "unit cost differs",
    }

    vendor_code = sa.Column(sa.String(length=30), nullable=True, doc="""
    Vendor's unique code for the product.  The meaning of this corresponds to that
    of the :attr:`ProductCost.code` column.
    """)

    case_quantity = sa.Column(sa.Integer(), nullable=False, default=1, doc="""
    Number of units in a case of product.
    """)

    ordered_cases = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of cases of the product which were originally ordered from the vendor.
    """)

    ordered_units = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of units of the product which were originally ordered from the vendor.
    """)

    shipped_cases = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of cases of the product which were shipped by the vendor.
    """)

    shipped_units = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of units of the product which were shipped by the vendor.
    """)

    out_of_stock = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether the item was known to be "out of stock" per the
    vendor invoice.
    """)

    case_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Cost per case of the product.
    """)

    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Cost per unit of the product.
    """)

    unit_cost_diff = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Unit cost difference between the invoice and product's current cost record.
    """)

    total_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Total cost for this product line item; should equate to the number of units
    shipped multiplied by the unit cost.
    """)

    line_number = sa.Column(sa.Integer(), nullable=True, doc="""
    Line number of the purchase order with which this invoice row matches.  Custom
    batch handlers may populate and leverage this field, but the default handler
    does not.
    """)



# -*- coding: utf-8 -*-
"""prevent null for batch.comlete

Revision ID: 42edf57f1ae2
Revises: 391fac830cb1
Create Date: 2018-09-22 18:58:20.563028

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '42edf57f1ae2'
down_revision = '391fac830cb1'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_handheld
    batch = sa.sql.table('batch_handheld', sa.sql.column('complete'))
    op.execute(batch.update()\
               .where(batch.c.complete == None)\
               .values({'complete': False}))
    op.alter_column('batch_handheld', 'complete', nullable=False)

    # batch_importer
    batch = sa.sql.table('batch_importer', sa.sql.column('complete'))
    op.execute(batch.update()\
               .where(batch.c.complete == None)\
               .values({'complete': False}))
    op.alter_column('batch_importer', 'complete', nullable=False)

    # batch_inventory
    batch = sa.sql.table('batch_inventory', sa.sql.column('complete'))
    op.execute(batch.update()\
               .where(batch.c.complete == None)\
               .values({'complete': False}))
    op.alter_column('batch_inventory', 'complete', nullable=False)

    # batch_pricing
    batch = sa.sql.table('batch_pricing', sa.sql.column('complete'))
    op.execute(batch.update()\
               .where(batch.c.complete == None)\
               .values({'complete': False}))
    op.alter_column('batch_pricing', 'complete', nullable=False)

    # label_batch
    batch = sa.sql.table('label_batch', sa.sql.column('complete'))
    op.execute(batch.update()\
               .where(batch.c.complete == None)\
               .values({'complete': False}))
    op.alter_column('label_batch', 'complete', nullable=False)

    # purchase_batch
    batch = sa.sql.table('purchase_batch', sa.sql.column('complete'))
    op.execute(batch.update()\
               .where(batch.c.complete == None)\
               .values({'complete': False}))
    op.alter_column('purchase_batch', 'complete', nullable=False)

    # vendor_catalog
    batch = sa.sql.table('vendor_catalog', sa.sql.column('complete'))
    op.execute(batch.update()\
               .where(batch.c.complete == None)\
               .values({'complete': False}))
    op.alter_column('vendor_catalog', 'complete', nullable=False)

    # vendor_invoice
    batch = sa.sql.table('vendor_invoice', sa.sql.column('complete'))
    op.execute(batch.update()\
               .where(batch.c.complete == None)\
               .values({'complete': False}))
    op.alter_column('vendor_invoice', 'complete', nullable=False)


def downgrade():

    # downgrade is simpler, we just need to allow null again
    op.alter_column('vendor_invoice', 'complete', nullable=True)
    op.alter_column('vendor_catalog', 'complete', nullable=True)
    op.alter_column('purchase_batch', 'complete', nullable=True)
    op.alter_column('label_batch', 'complete', nullable=True)
    op.alter_column('batch_pricing', 'complete', nullable=True)
    op.alter_column('batch_inventory', 'complete', nullable=True)
    op.alter_column('batch_importer', 'complete', nullable=True)
    op.alter_column('batch_handheld', 'complete', nullable=True)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor catalog parser for Lotus Light
"""

from __future__ import unicode_literals, absolute_import

import os
import re
import datetime
import logging

import xlrd

from rattail.db import model
from rattail.gpc import GPC
from rattail.vendors.catalogs import CatalogParser


log = logging.getLogger(__name__)


class LotusLightCatalogParser(CatalogParser):
    """
    Vendor catalog parser for Lotus Light.
    """
    key = 'rattail.contrib.lotuslight'
    display = "Lotus Light"
    vendor_key = 'lotus-light'

    def parse_effective_date(self, path):
        """
        This parser does not expect to find an effective date within the
        catalog file itself, but it makes a feeble attempt at gleaning it from
        the filename.  The example used to create this code was:

        .. code-block:: none

           Lotus Light Catalog Nov 2012.xls
        """
        basename = os.path.basename(path)
        match = re.match(r'^Lotus Light Catalog (\w{3} \d{4})\.xls$', basename)
        if match:
            return datetime.datetime.strptime(match.group(1), '%b %Y').date()

    def parse_rows(self, path, progress=None):
        """
        Parse data rows from a catalog file.
        """
        book = xlrd.open_workbook(path)
        sheet = book.sheet_by_index(0)
        for i in range(1, sheet.nrows): # skip first header row

            row = model.VendorCatalogRow()

            upc = sheet.cell_value(i, 1)
            if upc:
                if upc.isdigit():
                    # Sometimes their catalog doesn't include the check digit..?
                    if len(upc) == 11:
                        row.upc = GPC(int(upc), calc_check_digit='upc')
                    else:
                        row.upc = GPC(int(upc))
                else:
                    log.warning("invalid UPC at row {0}: {1}".format(i + 1, repr(upc)))

            row.brand_name = sheet.cell_value(i, 2)
            row.description = sheet.cell_value(i, 4)
            row.vendor_code = sheet.cell_value(i, 0)
            row.case_quantity = 1
            row.unit_cost = self.decimal(sheet.cell_value(i, 7))
            row.case_cost = row.unit_cost
            yield row



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Contact Info
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy import orm

from .core import Base, uuid_column


@six.python_2_unicode_compatible
class PhoneNumber(Base):
    """
    Represents a phone (or fax) number associated with a contactable entity.
    """
    __tablename__ = 'phone'
    __versioned__= {}

    uuid = uuid_column()
    parent_type = sa.Column(sa.String(length=20), nullable=False)
    parent_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)
    type = sa.Column(sa.String(length=15))
    number = sa.Column(sa.String(length=20), nullable=False)

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __str__(self):
        return self.number or ""


@six.python_2_unicode_compatible
class EmailAddress(Base):
    """
    Represents an email address associated with a contactable entity.
    """
    __tablename__ = 'email'
    __versioned__= {}

    uuid = uuid_column()
    parent_type = sa.Column(sa.String(length=20), nullable=False)
    parent_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)
    type = sa.Column(sa.String(length=15))
    address = sa.Column(sa.String(length=255), nullable=False)

    invalid = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether the email address is *known* to be invalid.
    Defaults to NULL, meaning the validity is "not known".
    """)

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __str__(self):
        return self.address or ""


@six.python_2_unicode_compatible
class MailingAddress(Base):
    """
    Represents a physical / mailing address associated with a contactable entity.
    """
    __tablename__ = 'address'
    __versioned__= {}

    uuid = uuid_column()
    parent_type = sa.Column(sa.String(length=20), nullable=False)
    parent_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)
    type = sa.Column(sa.String(length=15), nullable=True)

    street = sa.Column(sa.String(length=100), nullable=True)
    street2 = sa.Column(sa.String(length=100), nullable=True)
    city = sa.Column(sa.String(length=60), nullable=True)
    state = sa.Column(sa.String(length=2), nullable=True)
    zipcode = sa.Column(sa.String(length=10), nullable=True)
    invalid = sa.Column(sa.Boolean(), nullable=True)

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __str__(self):

        if self.street and self.street2:
            street = '{}, {}'.format(self.street, self.street2)
        else:
            street = self.street or ''

        if self.city and self.state:
            city = '{}, {}'.format(self.city, self.state)
        else:
            city = self.city or self.state or ''

        if street and city and self.zipcode:
            text = '{}, {}  {}'.format(street, city, self.zipcode)
        elif street and city:
            text = '{}, {}'.format(street, city)
        elif street and self.zipcode:
            text = '{}  {}'.format(street, self.zipcode)
        elif city and self.zipcode:
            text = '{}  {}'.format(city, self.zipcode)
        else:
            text = city or self.zipcode or ''

        return text


class ContactMixin(object):
    """
    Mixin which provides some useful methods for "contact" models, i.e. those
    which can play "parent" to email, phone and address records.
    """
    # TODO: subclass must set these
    _contact_email_model = None
    _contact_phone_model = None
    _contact_address_model = None

    def make_email(self, **kwargs):
        """
        Make a new "email" record for the contact.
        """
        email = self._contact_email_model(**kwargs)
        return email

    def add_email(self, **kwargs):
        """
        Add a new "email" record to the contact.
        """
        flush = kwargs.pop('flush', True)
        email = self.make_email(**kwargs)
        self.emails.append(email)
        if flush:
            session = orm.object_session(self)
            session.flush()
        return email

    def set_primary_email(self, email, flush=True):
        """
        Will re-arrange the contact's email records as needed to ensure that
        the given ``email`` record is "primary" - i.e. first in the list.
        """
        if email.preference != 1:
            session = orm.object_session(self)
            contact = email.parent
            if not contact:
                contact = session.query(email.Parent).get(email.parent_uuid)
                if not contact:
                    raise ValueError("cannot locate parent {} contact for email: {}".format(
                        email.Parent.__name__, email))
            emails = contact.emails
            if email in emails:
                emails.remove(email)
            emails.insert(0, email)
            emails.reorder()
            if flush:
                session.flush()

    def remove_email(self, email, **kwargs):
        """
        Remove the given email record from the contact.
        """
        flush = kwargs.pop('flush', True)
        self.emails.remove(email)
        if flush:
            session = orm.object_session(self)
            session.flush()

    def make_phone(self, **kwargs):
        """
        Make a new "phone" record for the contact.
        """
        # set some safe defaults in case session is flushed early
        kwargs.setdefault('number', '')
        phone = self._contact_phone_model(**kwargs)
        return phone

    def add_phone(self, **kwargs):
        """
        Add a new "phone" record to the contact.
        """
        flush = kwargs.pop('flush', True)
        phone = self.make_phone(**kwargs)
        self.phones.append(phone)
        if flush:
            session = orm.object_session(self)
            session.flush()
        return phone

    def remove_phone(self, phone, **kwargs):
        """
        Remove the given phone record from the contact.
        """
        flush = kwargs.pop('flush', True)
        self.phones.remove(phone)
        if flush:
            session = orm.object_session(self)
            session.flush()

    def set_primary_phone(self, phone, flush=True):
        """
        Will re-arrange the contact's phone records as needed to ensure that
        the given ``phone`` record is "primary" - i.e. first in the list.
        """
        if phone.preference != 1:
            session = orm.object_session(self)
            contact = phone.parent
            if not contact:
                contact = session.query(phone.Parent).get(phone.parent_uuid)
                if not contact:
                    raise ValueError("cannot locate parent {} contact for phone: {}".format(
                        phone.Parent.__name__, phone))
            phones = contact.phones
            if phone in phones:
                phones.remove(phone)
            phones.insert(0, phone)
            phones.reorder()
            if flush:
                session.flush()

    def make_address(self, **kwargs):
        """
        Make a new "address" record for the contact.
        """
        address = self._contact_address_model(**kwargs)
        return address

    def add_address(self, **kwargs):
        """
        Add a new "address" record to the contact.
        """
        flush = kwargs.pop('flush', True)
        address = self.make_address(**kwargs)
        self.addresses.append(address)
        if flush:
            session = orm.object_session(self)
            session.flush()
        return address

    def set_primary_address(self, address, flush=True):
        """
        Will re-arrange the contact's address records as needed to ensure that
        the given ``address`` record is "primary" - i.e. first in the list.
        """
        if address.preference != 1:
            session = orm.object_session(self)
            contact = address.parent
            if not contact:
                contact = session.query(address.Parent).get(address.parent_uuid)
                if not contact:
                    raise ValueError("cannot locate parent {} contact for address: {}".format(
                        address.Parent.__name__, address))
            addresses = contact.addresses
            if address in addresses:
                addresses.remove(address)
            addresses.insert(0, address)
            addresses.reorder()
            if flush:
                session.flush()

    def remove_address(self, address, **kwargs):
        """
        Remove the given address record from the contact.
        """
        flush = kwargs.pop('flush', True)
        self.addresses.remove(address)
        if flush:
            session = orm.object_session(self)
            session.flush()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Luigi utilities
"""

from __future__ import unicode_literals, absolute_import

import os
import datetime
import subprocess
import sys
import logging

import luigi


class OvernightTask(luigi.Task):
    """
    Base class for overnight automation tasks.
    """
    date = luigi.DateParameter()

    # TODO: subclass must define this
    filename = None

    # how long should we wait after task completes, for datasync to catch up?
    datasync_wait_minutes = None

    def output(self):
        return luigi.LocalTarget('{}/{}'.format(self.date.strftime('%Y/%m/%d'), self.filename))

    def run_command(self):
        raise NotImplementedError

    def touch_output(self):
        with self.output().open('w') as f:
            pass

    def datasync_wait(self, minutes=None):
        if minutes is None:
            minutes = self.datasync_wait_minutes or 10
        subprocess.check_call([
            'bin/rattail',
            '--config=app/cron.conf',
            'datasync',
            '--timeout={}'.format(minutes),
            'wait',
        ])

    def run(self):
        workdir = os.getcwd()
        os.chdir(sys.prefix)
        self.date_plus = self.date + datetime.timedelta(days=1)
        self.run_command()
        if self.datasync_wait_minutes:
            self.datasync_wait()
        os.chdir(workdir)
        self.touch_output()


class WarnSummaryIfProblems(logging.Filter):
    """
    Custom logging filter, to elevate the Luigi "execution summary" message to
    WARNING level, if any problems are detected for the run.  Note that this
    simply checks for the ``:)`` message to know if there were problem.
    """

    good_messages = [
        "This progress looks :) because there were no failed tasks or missing external dependencies",
        "This progress looks :) because there were no failed tasks or missing dependencies",
    ]


    def filter(self, record):
        if record.name == 'luigi-interface' and record.levelno == logging.INFO:
            if "===== Luigi Execution Summary =====" in record.msg:
                if not any([msg in record.msg for msg in self.good_messages]):
                    record.levelno = logging.WARNING
                    record.levelname = 'WARNING'
        return True



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Application Configuration
"""

from __future__ import unicode_literals, absolute_import

import os
import re
import sys
import shlex
import datetime
from six.moves import configparser
import warnings
import logging
import logging.config

import six

from rattail.util import load_entry_points, import_module_path
from rattail.exceptions import WindowsExtensionsNotInstalled, ConfigurationError
from rattail.files import temp_path
from rattail.logging import TimeConverter
from rattail.util import prettify


log = logging.getLogger(__name__)


def parse_bool(value):
    """
    Derive a boolean from the given string value.
    """
    if value is None:
        return None
    if isinstance(value, bool):
        return value
    if value.lower() in ('true', 'yes', 'on', '1'):
        return True
    return False


def parse_list(value):
    """
    Parse a configuration value, splitting by whitespace and/or commas and
    taking quoting into account, etc., yielding a list of strings.
    """
    if value is None:
        return []
    # Per the shlex docs (https://docs.python.org/2/library/shlex.html):
    # "Prior to Python 2.7.3, this module did not support Unicode input."
    if six.PY2 and isinstance(value, six.text_type):
        value = value.encode('utf-8')
    parser = shlex.shlex(value)
    parser.whitespace += ','
    parser.whitespace_split = True
    values = list(parser)
    for i, value in enumerate(values):
        if value.startswith('"') and value.endswith('"'):
            values[i] = value[1:-1]
        elif value.startswith("'") and value.endswith("'"):
            values[i] = value[1:-1]
    return values


class RattailConfig(object):
    """
    Rattail config object; this represents the sum total of configuration
    available to the running app.  The actual config available falls roughly
    into two categories: the "defaults" and the "db" (more on these below).
    The general idea here is that one might wish to provide some default
    settings within some config file(s) and/or the command line itself, but
    then allow all settings found in the database to override those defaults.
    However, all variations on this theme are supported, e.g. "use db settings
    but prefer those from file", "never use db settings", and so on.

    As for the "defaults" aspect of the config, this is read only once upon
    application startup.  It almost certainly involves one (or more) config
    file(s), but in addition to that, the application itself is free to embed
    default settings within the config object.  When this occurs, there will be
    no distinction made between settings which came from a file versus those
    which were established as defaults by the application logic.

    As for the "db" aspect of the config, of course this ultimately hinges upon
    the config defaults.  If a default Rattail database connection is defined,
    then the ``Setting`` table within that database may also be consulted for
    config values.  When this is done, the ``Setting.name`` is determined by
    concatenating the ``section`` and ``option`` arguments from the
    :meth:`get()` call, with a period (``'.'``) in between.
    """

    def __init__(self, files=[], usedb=None, preferdb=None):
        self.files_requested = []
        self.files_read = []
        if six.PY3:
            self.parser = configparser.ConfigParser()
        else: # PY2
            self.parser = configparser.SafeConfigParser()
        for path in files:
            self.read_file(path)
        self.usedb = usedb
        if self.usedb is None:
            self.usedb = self.getbool('rattail.config', 'usedb', usedb=False, default=False)
        self.preferdb = preferdb
        if self.usedb and self.preferdb is None:
            self.preferdb = self.getbool('rattail.config', 'preferdb', usedb=False, default=False)

        # Attempt to detect lack of SQLAlchemy libraries etc.  This allows us
        # to avoid installing those on a machine which will not need to access
        # a database etc.
        self._session_factory = None
        if self.usedb:
            try:
                from rattail.db import Session
            except ImportError: # pragma: no cover
                log.warning("config created with `usedb = True`, but can't import "
                            "`rattail.db.Session`, so setting `usedb = False` instead",
                            exc_info=True)
                self.usedeb = False
                self.preferdb = False
            else:
                self._session_factory = lambda: (Session(), True)

    def read_file(self, path, recurse=True):
        """
        Read in config from the given file.
        """
        path = os.path.abspath(path)
        if path in self.files_requested:
            log.debug("ignoring config file which was already requested: {0}".format(path))
            return

        log.debug("will attempt to read config from file: {0}".format(path))
        self.files_requested.append(path)

        if six.PY3:
            ConfigParser = configparser.ConfigParser
        else: # PY2
            ConfigParser = configparser.SafeConfigParser
        parser = ConfigParser(dict(
            here=os.path.dirname(path),
        ))
        if not parser.read(path):
            log.debug("ConfigParser.read() failed")
            return

        # If recursing, walk the complete config file inheritance chain.
        if recurse:
            if parser.has_section('rattail.config'):
                if parser.has_option('rattail.config', 'include'):
                    includes = parse_list(parser.get('rattail.config', 'include'))
                    for included in includes:
                        self.read_file(included, recurse=True)

        # Okay, now we can finally read this file into our main parser.
        self.parser.read(path)
        self.files_read.append(path)
        log.info("config was read from file: {0}".format(path))

    def configure_logging(self):
        """
        This first checks current config to determine whether or not we're
        supposed to be configuring logging at all.  If not, nothing more is
        done.

        If we are to configure logging, then this will save the current config
        parser defaults to a temporary file, and use this file to configure
        Python's standard logging module.
        """
        if not self.getbool('rattail.config', 'configure_logging', usedb=False, default=False):
            return

        # Coerce all logged timestamps to the local timezone, if possible.
        logging.Formatter.converter = TimeConverter(self)

        # Flush all current config to a single file, for input to fileConfig().
        path = temp_path(suffix='.conf')
        with open(path, 'wt') as f:
            self.parser.write(f)

        try:
            logging.config.fileConfig(path, disable_existing_loggers=False)
        except configparser.NoSectionError as error:
            log.warning("tried to configure logging, but got NoSectionError: {0}".format(error))
        else:
            log.debug("configured logging")
        finally:
            os.remove(path)

    def setdefault(self, section, option, value):
        """
        Establishes a new default for the given setting, if none exists yet.
        The effective default value is returned in all cases.
        """
        exists = True
        if not self.parser.has_section(section):
            self.parser.add_section(section)
            exists = False
        elif not self.parser.has_option(section, option):
            exists = False
        if not exists:
            self.parser.set(section, option, value)
        return self.parser.get(section, option)

    def set(self, section, option, value):
        """
        Set a value within the config's parser data set, i.e. the "defaults".
        This should probably be used sparingly, though one expected use is
        within tests (for convenience).
        """
        if not self.parser.has_section(section):
            self.parser.add_section(section)
        self.parser.set(section, option, value)

    def get(self, section, option, usedb=None, preferdb=None, session=None, default=None):
        """
        Retrieve a value from config.
        """
        usedb = usedb if usedb is not None else self.usedb
        if usedb:
            preferdb = preferdb if preferdb is not None else getattr(self, 'preferdb', False)
        else:
            preferdb = False
        
        if usedb and preferdb:
            value = self._getdb(section, option, session=session)
            if value is not None:
                return value

        if self.parser.has_option(section, option):
            return self.parser.get(section, option)

        if usedb and not preferdb:
            value = self._getdb(section, option, session=session)
            if value is not None:
                return value

        return default

    def _getdb(self, section, option, session=None):
        """
        Retrieve a config value from database settings table.
        """
        from rattail.db import api

        close = False
        if session is None:
            session, close = self._session_factory()
        value = api.get_setting(session, '{}.{}'.format(section, option))
        if close:
            session.close()
        return value

    def setdb(self, section, option, value, session=None):
        """
        Set a config value in the database settings table.  Note that the
        ``value`` arg should be a Unicode object.
        """
        from rattail.db import api

        close = False
        if session is None:
            session, close = self._session_factory()
        api.save_setting(session, '{}.{}'.format(section, option), value)
        if close:
            session.commit()
            session.close()

    def getbool(self, *args, **kwargs):
        """
        Retrieve a boolean value from config.
        """
        value = self.get(*args, **kwargs)
        return parse_bool(value)

    def getint(self, *args, **kwargs):
        """
        Retrieve an integer value from config.
        """
        value = self.get(*args, **kwargs)
        if value is None:
            return None
        if isinstance(value, int):
            return value
        return int(value)

    def getdate(self, *args, **kwargs):
        """
        Retrieve a date value from config.
        """
        value = self.get(*args, **kwargs)
        if value is None:
            return None
        if isinstance(value, datetime.date):
            return value
        return datetime.datetime.strptime(value, '%Y-%m-%d').date()

    def getlist(self, *args, **kwargs):
        """
        Retrieve a list of string values from a single config option.
        """
        value = self.get(*args, **kwargs)
        if value is None:
            return None
        if isinstance(value, six.string_types):
            return parse_list(value)
        return value            # maybe a caller-provided default?

    def get_dict(self, section):
        """
        Convenience method which returns a dictionary of options contained
        within the given section.  Note that this method only supports the
        "default" config settings, i.e. those within the underlying parser.
        """
        settings = {}
        if self.parser.has_section(section):
            for option in self.parser.options(section):
                settings[option] = self.parser.get(section, option)
        return settings

    def require(self, section, option, **kwargs):
        """
        Fetch a value from current config, and raise an error if no value can
        be found.
        """
        if 'default' in kwargs:
            warnings.warn("You have provided a default value to the `RattailConfig.require()` "
                          "method.  This is allowed but also somewhat pointless, since `get()` "
                          "would suffice if a default is known.", UserWarning)

        msg = kwargs.pop('msg', None)
        value = self.get(section, option, **kwargs)
        if value is not None:
            return value

        if msg is None:
            msg = "Missing or invalid config"
        msg = "{0}; please set '{1}' in the [{2}] section of your config file".format(
            msg, option, section)
        raise ConfigurationError(msg)

    ##############################
    # convenience methods
    ##############################

    def app_title(self, default=None):
        """
        Returns official display title for the current app.
        """
        if not default:
            return self.require('rattail', 'app_title')
        return self.get('rattail', 'app_title', default=default)

    def node_title(self, default=None):
        """
        Returns official display title for the current app.
        """
        title = self.get('rattail', 'node_title')
        if title:
            return title
        return self.app_title(default=default)

    def node_type(self, default=None):
        """
        Returns the "type" of current node.  What this means will generally
        depend on the app logic.
        """
        try:
            return self.require('rattail', 'node_type', usedb=False)
        except ConfigurationError:
            if default:
                return default
            raise

    def production(self):
        """
        Returns boolean indicating whether the app is running in production mode
        """
        return self.getbool('rattail', 'production', default=False)

    def demo(self):
        """
        Returns boolean indicating whether the app is running in demo mode
        """
        return self.getbool('rattail', 'demo', default=False)

    def versioning_enabled(self):
        """
        Returns boolean indicating whether data versioning is enabled.
        """
        return self.getbool('rattail.db', 'versioning.enabled', usedb=False,
                            default=False)

    def appdir(self, require=True):
        """
        Returns path to the 'app' dir, if known.
        """
        get = self.require if require else self.get
        return get('rattail', 'appdir')

    def datadir(self, require=True):
        """
        Returns path to the 'data' dir, if known.
        """
        get = self.require if require else self.get
        return get('rattail', 'datadir')

    def workdir(self, require=True):
        """
        Returns path to the 'work' dir, if known.
        """
        get = self.require if require else self.get
        return get('rattail', 'workdir')

    def batch_filedir(self, key=None):
        """
        Returns path to root folder where batches (optionally of type 'key')
        are stored.
        """
        path = os.path.abspath(self.require('rattail', 'batch.files'))
        if key:
            return os.path.join(path, key)
        return path

    def batch_filepath(self, key, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to a batch's data folder, with optional filename
        appended.  If ``makedirs`` is set, the batch data folder will be
        created if it does not already exist.
        """
        rootdir = self.batch_filedir(key)
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def export_filedir(self, key=None):
        """
        Returns path to root folder where exports (optionally of type 'key')
        are stored.
        """
        path = os.path.abspath(self.require('rattail', 'export.files'))
        if key:
            return os.path.join(path, key)
        return path

    def export_filepath(self, key, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to export data file, generated from the given args.
        """
        rootdir = self.export_filedir(key)
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def upgrade_filedir(self):
        """
        Returns path to root folder where upgrade files are stored.
        """
        path = os.path.abspath(self.require('rattail.upgrades', 'files'))
        return path

    def upgrade_filepath(self, uuid, filename=None, makedirs=False):
        """
        Returns absolute path to upgrade data file, generated from the given args.
        """
        rootdir = self.upgrade_filedir()
        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
        if makedirs and not os.path.exists(filedir):
            os.makedirs(filedir)
        if filename:
            return os.path.join(filedir, filename)
        return filedir

    def upgrade_command(self, default='/bin/sleep 30'):
        """
        Returns command to be used when performing upgrades.
        """
        # NOTE: we don't allow command to be specified in DB, for security reasons..
        return self.getlist('rattail.upgrades', 'command', default=default, usedb=False)

    def datasync_url(self):
        """
        Returns configured URL for managing datasync daemon.
        """
        return self.get('rattail.datasync', 'url')

    def get_enum(self, **kwargs):
        """
        Returns a reference to configured 'enum' module; defaults to
        :mod:`rattail.enum`.
        """
        kwargs.setdefault('usedb', False)
        spec = self.get('rattail', 'enum', default='rattail.enum', **kwargs)
        return import_module_path(spec)

    def get_model(self):
        """
        Returns a reference to configured 'model' module; defaults to
        :mod:`rattail.db.model`.
        """
        spec = self.get('rattail', 'model', default='rattail.db.model', usedb=False)
        return import_module_path(spec)

    def get_trainwreck_model(self):
        """
        Returns a reference to the configured data 'model' module for
        Trainwreck.  Note that there is *not* a default value for this; it must
        be configured.
        """
        spec = self.require('rattail.trainwreck', 'model', usedb=False)
        return import_module_path(spec)

    def product_key(self):
        """
        Returns the name of the attribute which should be treated as the
        canonical product key field, e.g. 'upc' or 'item_id' etc.
        """
        return self.get('rattail', 'product.key', default='upc')

    def product_key_title(self, key=None):
        """
        Returns the title string to be used when displaying product key field,
        e.g. "UPC" or "Part No." etc.
        """
        title = self.get('rattail', 'product.key_title')
        if title:
            return title
        if not key:
            key = self.product_key()
        if key == 'upc':
            return "UPC"
        if key == 'item_id':
            return "Item ID"
        return prettify(key)

    def single_store(self):
        """
        Returns boolean indicating whether the system is configured to behave
        as if it belongs to a single Store.
        """
        return self.getbool('rattail', 'single_store', default=False)

    def get_store(self, session):
        """
        Returns a :class:`rattail.db.model.Store` instance corresponding to app
        config, or ``None``.
        """
        from rattail.db import api

        store = self.get('rattail', 'store')
        if store:
            return api.get_store(session, store)


    ##############################
    # deprecated methods
    ##############################

    def options(self, section):
        warnings.warn("RattailConfig.option() is deprecated, please find "
                      "another way to accomplish what you're after.",
                      DeprecationWarning)
        return self.parser.options(section)

    def has_option(self, section, option):
        warnings.warn("RattailConfig.has_option() is deprecated, please find "
                      "another way to accomplish what you're after.",
                      DeprecationWarning)
        return self.parser.has_option(section, option)


class ConfigExtension(object):
    """
    Base class for all config extensions.
    """
    key = None

    def __repr__(self):
        return "ConfigExtension(key={0})".format(repr(self.key))

    def configure(self, config):
        """
        All subclasses should override this method, to extend the config object
        in any way necessary etc.
        """


def make_config(files=None, usedb=None, preferdb=None, env=None, winsvc=None, extend=True, versioning=None):
    """
    Returns a new config object, initialized with the given parameters and
    further modified by all registered config extensions.

    :param versioning: Controls whether or not the versioning system is
       configured with the new config object.  If ``True``, versioning will be
       configured.  If ``False`` then it will not be configured.  If ``None``
       (the default) then versioning will be configured only if the config
       object itself says that it should be.
    """
    if env is None:
        env = os.environ
    if files is None:
        files = env.get('RATTAIL_CONFIG_FILES')
        if files is not None:
            files = files.split(os.pathsep)
        else:
            files = default_system_paths() + default_user_paths()
    elif isinstance(files, six.string_types):
        files = [files]

    # If making config for a Windows service, we must read the default config
    # file(s) first, and check it to see if there is an alternate config file
    # which should be considered the "root" file.  Normally we specify the root
    # config file(s) via command line etc., but there is no practical way to
    # pass parameters to a Windows service.  This way we can effectively do
    # just that, via config.
    if winsvc is not None:
        parser = configparser.SafeConfigParser()
        parser.read(files)
        if parser.has_section('rattail.config'):
            key = 'winsvc.{0}'.format(winsvc)
            if parser.has_option('rattail.config', key):
                files = parse_list(parser.get('rattail.config', key))

    # Initial config object will have values read from the given file paths,
    # and kwargs, but no other app defaults etc. will have been applied yet.
    config = RattailConfig(files, usedb=usedb, preferdb=preferdb)
    config.configure_logging()

    if config.getbool('rattail', 'suppress_psycopg2_wheel_warning', usedb=False):
        # TODO: revisit this, does it require action from us?
        # suppress this warning about psycopg2 wheel; not sure what it means yet
        # exactly but it's causing frequent noise for us...
        warnings.filterwarnings(
            'ignore',
            r'^The psycopg2 wheel package will be renamed from release 2\.8; in order to keep '
            'installing from binary please use "pip install psycopg2-binary" instead\. For details '
            'see: <http://initd.org/psycopg/docs/install.html#binary-install-from-pypi>\.',
            UserWarning,
            r'^psycopg2$',
        )

    # Apply extra config for all available extensions.
    if extend:
        extensions = load_entry_points('rattail.config.extensions')
        for extension in extensions.values():
            log.debug("applying '{0}' config extension".format(extension.key))
            extension().configure(config)

    # maybe configure versioning
    if versioning is None:
        versioning = config.versioning_enabled()
    if versioning:
        from rattail.db.config import configure_versioning
        configure_versioning(config)

    return config


def default_system_paths():
    """
    Returns a list of default system-level config file paths, according to the
    current platform.
    """
    if sys.platform == 'win32':

        # Use the Windows Extensions libraries to fetch official defaults.
        try:
            from win32com.shell import shell, shellcon
        except ImportError:
            raise WindowsExtensionsNotInstalled
        else:
            return [
                os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_COMMON_APPDATA), 'rattail.conf'),
                os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_COMMON_APPDATA), 'rattail', 'rattail.conf'),
            ]

    return [
        '/etc/rattail.conf',
        '/etc/rattail/rattail.conf',
        '/usr/local/etc/rattail.conf',
        '/usr/local/etc/rattail/rattail.conf',
    ]


def default_user_paths():
    """
    Returns a list of default user-level config file paths, according to the
    current platform.
    """
    if sys.platform == 'win32':

        # Use the Windows Extensions libraries to fetch official defaults.
        try:
            from win32com.shell import shell, shellcon
        except ImportError:
            raise WindowsExtensionsNotInstalled
        else:
            return [
                os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_APPDATA), 'rattail.conf'),
                os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_APPDATA), 'rattail', 'rattail.conf'),
            ]

    return [
        os.path.expanduser('~/.rattail.conf'),
        os.path.expanduser('~/.rattail/rattail.conf'),
    ]


def get_user_dir(create=False):
    """
    Returns a path to the "preferred" user-level folder, in which additional
    config files (etc.) may be placed as needed.  This essentially returns a
    platform-specific variation of ``~/.rattail/``.

    If ``create`` is ``True``, then the folder will be created if it does not
    already exist.
    """
    if sys.platform == 'win32':

        # Use the Windows Extensions libraries to fetch official defaults.
        try:
            from win32com.shell import shell, shellcon
        except ImportError:
            raise WindowsExtensionsNotInstalled
        else:
            path = os.path.join(shell.SHGetSpecialFolderPath(
                0, shellcon.CSIDL_APPDATA), 'rattail')

    else:
        path = os.path.expanduser('~/.rattail')

    if create and not os.path.exists(path):
        os.mkdir(path)
    return path


def get_user_file(filename, createdir=False):
    """
    Returns a full path to a user-level config file location.  This is obtained
    by first calling :func:`get_user_dir()` and then joining the result with
    ``filename``.

    The ``createdir`` argument will be passed to :func:`get_user_dir()` as its
    ``create`` arg, and may be used to ensure the user-level folder exists.
    """
    return os.path.join(get_user_dir(create=createdir), filename)


class ConfigProfile(object):
    """
    Generic class to represent a config "profile", as used by the filemon and
    datasync daemons, etc.

    .. todo::
       This clearly needs more documentation.
    """

    @property
    def section(self):
        """
        Each subclass of ``ConfigProfile`` must define this.
        """
        raise NotImplementedError

    def _config_string(self, option, **kwargs):
        return self.config.get(self.section, '{0}.{1}'.format(self.key, option), **kwargs)

    def _config_boolean(self, option, default=None):
        return self.config.getbool(self.section, '{0}.{1}'.format(self.key, option),
                                   default=default)

    def _config_int(self, option, minimum=1, default=None):
        """
        Retrieve the *integer* value for the given option.
        """
        option = '{}.{}'.format(self.key, option)

        # try to read value from config
        value = self.config.getint(self.section, option)
        if value is not None:

            # found a value; validate it
            if value < minimum:
                log.warning("config value %s is too small; falling back to minimum "
                            "of %s for option: %s", value, minimum, option)
                value = minimum

        # or, use default value, if valid
        elif default is not None and default >= minimum:
            value = default

        # or, just use minimum value
        else:
            value = minimum

        return value

    def _config_list(self, option):
        return parse_list(self._config_string(option))


class FreeTDSLoggingFilter(logging.Filter):
    """
    Custom logging filter, to suppress certain "write to server failed"
    messages relating to FreeTDS database connections.  They seem harmless and
    just cause unwanted error emails.
    """

    def __init__(self, *args, **kwargs):
        logging.Filter.__init__(self, *args, **kwargs)
        self.pattern = re.compile(r'(?:Read from|Write to) the server failed')

    def filter(self, record):
        if (record.name == 'sqlalchemy.pool.QueuePool'
            and record.funcName == '_finalize_fairy'
            and record.levelno == logging.ERROR
            and record.msg == "Exception during reset or similar"
            and record.exc_info
            and self.pattern.search(six.text_type(record.exc_info[1]))):

            # Log this as a warning instead of error, to cut down on our noise.
            record.levelno = logging.WARNING
            record.levelname = 'WARNING'

        return True



# -*- coding: utf-8 -*-
"""enlarge product_cost.code

Revision ID: 571c070b7080
Revises: f950be5601a
Create Date: 2015-02-11 19:11:36.664874

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '571c070b7080'
down_revision = u'f950be5601a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():
    op.alter_column('product_cost', 'code', type_=sa.String(length=20))
    op.alter_column('product_cost_version', 'code', type_=sa.String(length=20))


def downgrade():
    op.alter_column('product_cost_version', 'code', type_=sa.String(length=15))
    op.alter_column('product_cost', 'code', type_=sa.String(length=15))



# -*- coding: utf-8 -*-
"""add batch_row.modified

Revision ID: 830a20917c49
Revises: d67dee1cb849
Create Date: 2018-07-10 08:58:35.083616

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '830a20917c49'
down_revision = u'd67dee1cb849'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_row
    op.add_column('batch_handheld_row', sa.Column('modified', sa.DateTime(), nullable=True))
    op.add_column('batch_inventory_row', sa.Column('modified', sa.DateTime(), nullable=True))
    op.add_column('batch_pricing_row', sa.Column('modified', sa.DateTime(), nullable=True))
    op.add_column('label_batch_row', sa.Column('modified', sa.DateTime(), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('modified', sa.DateTime(), nullable=True))
    op.add_column('vendor_catalog_row', sa.Column('modified', sa.DateTime(), nullable=True))
    op.add_column('vendor_invoice_row', sa.Column('modified', sa.DateTime(), nullable=True))


def downgrade():

    # batch_row
    op.drop_column('vendor_invoice_row', 'modified')
    op.drop_column('vendor_catalog_row', 'modified')
    op.drop_column('purchase_batch_row', 'modified')
    op.drop_column('label_batch_row', 'modified')
    op.drop_column('batch_pricing_row', 'modified')
    op.drop_column('batch_inventory_row', 'modified')
    op.drop_column('batch_handheld_row', 'modified')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail Data Exporters
"""

from __future__ import unicode_literals, absolute_import

import os
import csv

import six
from sqlalchemy import orm
from sqlalchemy_utils.functions import get_primary_keys

from rattail import csvutil
from rattail.util import OrderedDict
from rattail.db.util import make_topo_sortkey
from rattail.importing import Importer
from rattail.importing.rattail import FromRattailHandler, FromRattail
from rattail.importing.handlers import ToCSVHandler


class FromSQLAlchemyToCSVMixin(object):

    # subclass must define this
    FromParent = None

    def get_importers(self):
        """
        Here we build the set of available importers on the fly.  This avoids
        having to define things over and over since really we're just going to
        piggy-back on the existing Rattail -> Rattail logic, for obtaining the
        source data.
        """
        importers = {}
        model = self.get_model()

        # mostly try to make an importer for every data model
        for name in dir(model):
            obj = getattr(model, name, None)
            if isinstance(obj, type) and issubclass(obj, model.Base) and obj is not model.Base:
                importers[name] = self.make_importer_factory(name, obj)

        # sort importers according to topography
        topo_sortkey = make_topo_sortkey(model)
        importers = OrderedDict([
            (name, importers[name])
            for name in sorted(importers, key=topo_sortkey)
        ])

        return importers

    def make_importer_factory(self, name, cls):
        mapper = orm.class_mapper(cls)
        fields = list(mapper.columns.keys())
        pkeys = get_primary_keys(cls)
        return type('{}Importer'.format(name), (self.FromParent, ToCSV), {
            'model_class': cls,
            'supported_fields': fields,
            'simple_fields': fields,
            'key': list(pkeys),
        })


class FromRattailToCSV(FromSQLAlchemyToCSVMixin, FromRattailHandler, ToCSVHandler):
    """
    Handler for Rattail -> CSV data export.
    """
    direction = 'export'
    local_title = "CSV"
    FromParent = FromRattail

    @property
    def host_title(self):
        return self.config.node_title()

    def get_model(self):
        if self.config:
            return self.config.get_model()

        from rattail.db import model
        return model


class ToFile(Importer):
    """
    Base class for importers which target data file on the local side.
    """
    empty_local_data = True

    def setup(self):
        super(ToFile, self).setup()

        if not hasattr(self, 'output_file_path'):
            self.output_file_path = os.path.join(self.output_dir,
                                                 '{}.csv'.format(self.model_name))
        self.open_output_file()

    def teardown(self):
        self.close_output_file()

    def open_output_file(self):
        raise NotImplementedError

    def close_output_file(self):
        raise NotImplementedError

    def create_object(self, key, data):
        """
        Only "add" file record for object if not marked as "deleted".

        Note that you probably shouldn't need to override this method, but you
        probably do want to define :meth:`update_object()` instead.
        """
        if not data.get('_deleted_', False):
            return self.update_object(None, data)


class ToCSV(ToFile):
    """
    Base class for importers which target CSV file on the local side.
    """
    empty_local_data = True

    def open_output_file(self):
        if six.PY2:
            self.output_file = open(self.output_file_path, 'wb')
            self.output_writer = csvutil.UnicodeDictWriter(self.output_file, self.fields, encoding='utf_8')
        else: # PY3
            self.output_file = open(self.output_file_path, 'wt', encoding='utf_8')
            self.output_writer = csv.DictWriter(self.output_file, self.fields)
        self.output_writer.writeheader()

    def close_output_file(self):
        self.output_file.close()

    def update_object(self, obj, data, local_data=None):
        """
        Add objects's record to CSV output file.
        """
        data = self.coerce_csv(data)
        self.output_writer.writerow(data)
        return data

    def coerce_csv(self, data):
        """
        Coerce all field values in ``data`` to string, for CSV.
        """
        coerced = {}
        for field in self.fields:
            value = data[field]

            if value is None:
                value = ''

            else:
                value = six.text_type(value)

            coerced[field] = value
        return coerced



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Models for purchase order batches
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import (Base, BatchMixin, BatchRowMixin,
                              PurchaseBase, PurchaseItemBase, PurchaseCreditBase,
                              Purchase, PurchaseItem)
from rattail.db.core import uuid_column, filename_column


class PurchaseBatch(BatchMixin, PurchaseBase, Base):
    """
    Hopefully generic batch used for entering new purchases into the system, etc.?
    """
    batch_key = 'purchase'
    __tablename__ = 'purchase_batch'
    __batchrow_class__ = 'PurchaseBatchRow'
    model_title = "Purchasing Batch"
    model_title_plural = "Purchasing Batches"

    @declared_attr
    def __table_args__(cls):
        return cls.__batch_table_args__() + cls.__purchase_table_args__() + (
            sa.ForeignKeyConstraint(['purchase_uuid'], ['purchase.uuid'], name='purchase_batch_fk_purchase'),
            sa.ForeignKeyConstraint(['truck_dump_batch_uuid'], ['purchase_batch.uuid'], name='purchase_batch_fk_truck_dump_batch', use_alter=True),
        )

    STATUS_OK                   = 1
    STATUS_UNKNOWN_PRODUCT      = 2
    STATUS_TRUCKDUMP_UNCLAIMED  = 3
    STATUS_TRUCKDUMP_CLAIMED    = 4
    STATUS_UNKNOWN_COSTS        = 5

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_UNKNOWN_PRODUCT          : "has unknown product(s)",
        STATUS_UNKNOWN_COSTS            : "has unknown product cost(s)",
        STATUS_TRUCKDUMP_UNCLAIMED      : "not yet fully claimed",
        STATUS_TRUCKDUMP_CLAIMED        : "fully claimed by child(ren)",
    }

    purchase_uuid = sa.Column(sa.String(length=32), nullable=True)

    purchase = orm.relationship(
        Purchase,
        doc="""
        Reference to the purchase with which the batch is associated.  May be
        null, e.g. in the case of a "new purchase" batch.
        """,
        backref=orm.backref(
            'batches',
            order_by='PurchaseBatch.id',
            doc="""
            List of batches associated with the purchase.
            """))

    mode = sa.Column(sa.Integer(), nullable=False, doc="""
    Numeric "mode" for the purchase batch, to indicate new/receiving etc.
    """)

    invoice_file = filename_column(doc="Base name for the associated invoice file, if any.")

    invoice_parser_key = sa.Column(sa.String(length=100), nullable=True, doc="""
    The key of the parser used to read the contents of the invoice file.
    """)

    order_quantities_known = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether the order quantities were known at time of batch
    creation / population.  Really this is only used for batches of 'receiving'
    mode, to present a slightly different UI if order quantities were (not) known.
    """)

    po_total_calculated = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Calculated PO total, per quantity received thus far (if applicable).
    """)

    invoice_total_calculated = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Calculated invoice total, per quantity received thus far (if applicable).
    """)

    truck_dump = sa.Column(sa.Boolean(), nullable=True, default=False, doc="""
    Flag indicating whether a "receiving" batch is of the "truck dump"
    persuasion, i.e.  it does not correspond to a single purchase order but
    rather is assumed to represent multiple orders.
    """)

    truck_dump_children_first = sa.Column(sa.Boolean(), nullable=True, default=False, doc="""
    If batch is a "truck dump parent", this flag indicates whether its
    "children" are to be attached *first* as opposed to *last*.  If the flag is
    true, all children must be attached prior to the receiving process.  If
    flag is false, receiving must happen first, and then all children attached
    at the end.
    """)

    truck_dump_ready = sa.Column(sa.Boolean(), nullable=True, default=False, doc="""
    If batch is a "truck dump parent", this flag indicates whether it is
    "ready" for the actual receiving process.  If children are to be attached
    first, this flag should not be set until all children are attached.  If
    children are last, this flag should be set immediately upon batch creation.
    """)

    truck_dump_status = sa.Column(sa.Integer(), nullable=True, doc="""
    Truck dump status code for the row.  Only relevant for truck dump parent
    batches.  Indicates whether this parent has been fully claimed by children
    yet, etc.
    """)

    receiving_complete = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag to indicate whether the "receiving" process is complete for the batch.
    Not applicable to all batches of course; depends on handler logic etc.
    """)

    truck_dump_batch_uuid = sa.Column(sa.String(length=32), nullable=True)
    truck_dump_batch = orm.relationship(
        'PurchaseBatch',
        remote_side='PurchaseBatch.uuid',
        doc="""
        Reference to the "truck dump" receiving batch, for which the current
        batch represents a single invoice which partially "consumes" the truck
        dump.
        """,
        backref=orm.backref(
            'truck_dump_children',
            order_by='PurchaseBatch.id',
            doc="""
            List of batches which are "children" of the current batch, which is
            assumed to be a truck dump.
            """))

    def is_truck_dump_parent(self):
        """
        Returns boolean indicating whether or not the batch is a "truck dump"
        parent.
        """
        if self.truck_dump:
            return True
        return False

    def is_truck_dump_child(self):
        """
        Returns boolean indicating whether or not the batch is a "truck dump"
        child.
        """
        if self.truck_dump_batch:
            return True
        return False

    def is_truck_dump_related(self):
        """
        Returns boolean indicating whether or not the batch is associated with
        a "truck dump" in any way, i.e. is a parent or child of such.
        """
        if self.is_truck_dump_parent():
            return True
        if self.is_truck_dump_child():
            return True
        return False


class PurchaseBatchRow(BatchRowMixin, PurchaseItemBase, Base):
    """
    Row of data within a purchase batch.
    """
    __tablename__ = 'purchase_batch_row'
    __batch_class__ = PurchaseBatch

    @declared_attr
    def __table_args__(cls):
        return cls.__batchrow_table_args__() + cls.__purchaseitem_table_args__() + (
            sa.ForeignKeyConstraint(['item_uuid'], ['purchase_item.uuid'], name='purchase_batch_row_fk_item'),
        )

    STATUS_OK                           = 1
    STATUS_PRODUCT_NOT_FOUND            = 2
    STATUS_COST_NOT_FOUND               = 3
    STATUS_CASE_QUANTITY_UNKNOWN        = 4
    STATUS_INCOMPLETE                   = 5
    STATUS_ORDERED_RECEIVED_DIFFER      = 6
    STATUS_TRUCKDUMP_UNCLAIMED          = 7
    STATUS_TRUCKDUMP_CLAIMED            = 8
    STATUS_TRUCKDUMP_OVERCLAIMED        = 9
    STATUS_CASE_QUANTITY_DIFFERS        = 10
    STATUS_TRUCKDUMP_PARTCLAIMED        = 11
    STATUS_OUT_OF_STOCK                 = 12

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_PRODUCT_NOT_FOUND        : "product not found",
        STATUS_COST_NOT_FOUND           : "product found but not cost",
        STATUS_CASE_QUANTITY_UNKNOWN    : "case quantity not known",
        STATUS_CASE_QUANTITY_DIFFERS    : "case quantity differs",
        STATUS_INCOMPLETE               : "incomplete",
        STATUS_ORDERED_RECEIVED_DIFFER  : "shipped / received differ",
        STATUS_TRUCKDUMP_UNCLAIMED      : "not claimed by any child(ren)",
        STATUS_TRUCKDUMP_PARTCLAIMED    : "partially claimed by child(ren)",
        STATUS_TRUCKDUMP_CLAIMED        : "fully claimed by child(ren)",
        STATUS_TRUCKDUMP_OVERCLAIMED    : "OVER claimed by child(ren)",
        STATUS_OUT_OF_STOCK             : "out of stock",
    }

    item_entry = sa.Column(sa.String(length=20), nullable=True, doc="""
    Raw entry value, as obtained from the initial data source, and which is
    used to locate the product within the system.  This raw value is preserved
    in case the initial lookup fails and a refresh must attempt further
    lookup(s) later.  Only used by certain batch handlers in practice.
    """)

    item_uuid = sa.Column(sa.String(length=32), nullable=True)

    item = orm.relationship(
        PurchaseItem,
        doc="""
        Reference to the purchase item with which the batch row is associated.
        May be null, e.g. in the case of a "new purchase" batch.
        """)

    po_total_calculated = sa.Column(sa.Numeric(precision=7, scale=2), nullable=True, doc="""
    Calculated total cost for line item, per quantity received thus far.  See
    also :attr:`po_total`.
    """)

    catalog_cost_confirmed = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating that the "catalog" cost for this row has been confirmed,
    e.g. by a user.
    """)

    invoice_cost_confirmed = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating that the invoice cost for this row has been "confirmed",
    e.g. by a user.
    """)

    invoice_total_calculated = sa.Column(sa.Numeric(precision=7, scale=2), nullable=True, doc="""
    Calculated total cost for line item, per quantity received thus far.  See
    also :attr:`invoice_total`.
    """)

    truck_dump_status = sa.Column(sa.Integer(), nullable=True, doc="""
    Truck dump status code for the row.  Only relevant for truck dump parent
    batches.  Indicates whether this parent row has been fully claimed by
    children yet, etc.
    """)


class PurchaseBatchRowClaim(Base):
    """
    Represents the connection between a row(s) from a truck dump batch, and the
    corresponding "child" batch row which claims it, as well as the claimed
    quantities etc.
    """
    __tablename__ = 'purchase_batch_row_claim'
    __table_args__ = (
        sa.ForeignKeyConstraint(['claiming_row_uuid'], ['purchase_batch_row.uuid'], name='purchase_batch_row_claim_fk_claiming_row'),
        sa.ForeignKeyConstraint(['claimed_row_uuid'], ['purchase_batch_row.uuid'], name='purchase_batch_row_claim_fk_claimed_row'),
    )

    uuid = uuid_column()

    claiming_row_uuid = sa.Column(sa.String(length=32), nullable=False)
    claiming_row = orm.relationship(
        PurchaseBatchRow,
        foreign_keys='PurchaseBatchRowClaim.claiming_row_uuid',
        doc="""
        Reference to the "child" row which is claiming some row from a truck
        dump batch.
        """,
        backref=orm.backref(
            'truck_dump_claims',
            cascade='all, delete-orphan',
            doc="""
            List of claims which this "child" row makes against rows within a
            truck dump batch.
            """))

    claimed_row_uuid = sa.Column(sa.String(length=32), nullable=False)
    claimed_row = orm.relationship(
        PurchaseBatchRow,
        foreign_keys='PurchaseBatchRowClaim.claimed_row_uuid',
        doc="""
        Reference to the truck dump batch row which is claimed by the "child" row.
        """,
        backref=orm.backref(
            'claims',
            cascade='all, delete-orphan',
            doc="""
            List of claims made by "child" rows against this truck dump batch row.
            """))

    cases_received = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Number of cases of product which were ultimately received, and are involved in the claim.
    """)

    units_received = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Number of units of product which were ultimately received, and are involved in the claim.
    """)

    cases_damaged = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Number of cases of product which were shipped damaged, and are involved in the claim.
    """)

    units_damaged = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Number of units of product which were shipped damaged, and are involved in the claim.
    """)

    cases_expired = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Number of cases of product which were shipped expired, and are involved in the claim.
    """)

    units_expired = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Number of units of product which were shipped expired, and are involved in the claim.
    """)

    # TODO: should have fields for mispick here too, right?

    def is_empty(self):
        """
        Returns boolean indicating whether the claim is "empty" - i.e. if it
        has zero or null for all of its quantity fields.
        """
        if self.cases_received:
            return False
        if self.units_received:
            return False

        if self.cases_damaged:
            return False
        if self.units_damaged:
            return False

        if self.cases_expired:
            return False
        if self.units_expired:
            return False

        return True


class PurchaseBatchCredit(PurchaseCreditBase, Base):
    """
    Represents a working copy of purchase credit tied to a batch row.
    """
    __tablename__ = 'purchase_batch_credit'

    @declared_attr
    def __table_args__(cls):
        return cls.__purchasecredit_table_args__() + (
            sa.ForeignKeyConstraint(['row_uuid'], ['purchase_batch_row.uuid'], name='purchase_batch_credit_fk_row'),
        )

    uuid = uuid_column()

    row_uuid = sa.Column(sa.String(length=32), nullable=True)

    row = orm.relationship(
        PurchaseBatchRow,
        doc="""
        Reference to the batch row with which the credit is associated.
        """,
        backref=orm.backref(
            'credits',
            cascade='all, delete-orphan',
            doc="""
            List of :class:`PurchaseBatchCredit` instances for the row.
            """))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Problem Reports Framework
"""

from __future__ import unicode_literals, absolute_import

from rattail.util import progress_loop


class ProblemReport(object):
    """
    Base class for all problem reports.

    A "problem report" will generally contain logic which looks for problems of
    some kind or another, within the data of a given system.  It may or may not
    be able to "fix" some of these problems.  In either case it is expected to
    be able to generate and send an email report with the details.
    """
    system_key = None
    problem_key = None
    problem_title = None
    email_key = None

    def __init__(self, config, dry_run=False, progress=None):
        self.config = config
        self.dry_run = dry_run
        self.progress = progress
        self.enum = self.config.get_enum()

    def progress_loop(self, func, items, factory=None, **kwargs):
        factory = factory or self.progress
        return progress_loop(func, items, factory, **kwargs)

    def find_problems(self, **kwargs):
        """
        Find all problems for which this report is responsible.  This should
        always return a list, although no constraint is made on what the
        elements must be.
        """
        return []

    def get_email_context(self, problems, **kwargs):
        """
        This method can be used to add extra context for a specific report's
        email template.
        """
        return kwargs



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Trainwreck importing - shared utilities
"""

from __future__ import unicode_literals, absolute_import

import datetime

from rattail.time import localtime, make_utc


class ToOrFromTrainwreck(object):
    """
    Common base class for ToTrainwreck and FromTrainwreck base clases.
    """

    def setup(self):
        super(ToOrFromTrainwreck, self).setup()

        # start_time defaults to yesterday midnight
        if not hasattr(self, 'start_time'):
            if hasattr(self, 'start_date'):
                start_date = self.start_date
            elif self.args.start_date:
                start_date = self.args.start_date
            else:
                start_date = localtime(self.config).date() - datetime.timedelta(days=1)
            start_time = datetime.datetime.combine(start_date, datetime.time(0))
            self.start_time = localtime(self.config, start_time)

        # end_time defaults to now
        if not hasattr(self, 'end_time'):
            if hasattr(self, 'end_date'):
                end_date = self.end_date + datetime.timedelta(days=1)
                end_time = datetime.datetime.combine(end_date, datetime.time(0))
            elif self.args.end_date:
                end_date = self.args.end_date + datetime.timedelta(days=1)
                end_time = datetime.datetime.combine(end_date, datetime.time(0))
            else:
                end_time = None     # i.e. "now"
            self.end_time = localtime(self.config, end_time)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Core batch data models

Actually the classes in this module are not true models but rather are mixins,
which provide the common columns etc. for batch tables.
"""

from __future__ import unicode_literals, absolute_import

import os
import datetime
import shutil

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import relationship, object_session
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.orderinglist import ordering_list

from rattail.db.core import uuid_column, filename_column
from rattail.db.types import GPCType, JSONTextDict
from rattail.db.model import User, Product
from rattail.time import make_utc


@six.python_2_unicode_compatible
class BatchMixin(object):
    """
    Mixin for all (new-style) batch classes.
    """

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__()

    @classmethod
    def __batch_table_args__(cls):
        return (
            sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'],
                                    name='{0}_fk_created_by'.format(cls.__tablename__)),
            sa.ForeignKeyConstraint(['cognized_by_uuid'], ['user.uuid'],
                                    name='{0}_fk_cognized_by'.format(cls.__tablename__)), 
            sa.ForeignKeyConstraint(['executed_by_uuid'], ['user.uuid'],
                                    name='{0}_fk_executed_by'.format(cls.__tablename__)),
        )

    @classmethod
    def __default_table_args__(cls):
        return cls.__batch_table_args__()

    @declared_attr
    def batch_key(cls):
        return cls.__tablename__

    uuid = uuid_column()

    id = sa.Column(sa.Integer(), sa.Sequence('batch_id_seq'), nullable=False, doc="""
    Numeric ID for the batch, unique across all "new-style" batches within the
    Rattail database.
    """)

    description = sa.Column(sa.String(length=255), nullable=True, doc="""
    Basic (loosely identifying) description for the batch.
    """)

    created = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
    Date and time when the batch was first created.
    """)

    created_by_uuid = sa.Column(sa.String(length=32), nullable=False)

    @declared_attr
    def created_by(cls):
        return orm.relationship(
            User,
            primaryjoin=lambda: User.uuid == cls.created_by_uuid,
            foreign_keys=lambda: [cls.created_by_uuid],
            doc="""
            Reference to the :class:`User` who first created the batch.
            """)

    cognized = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date and time when the batch data was last cognized.
    """)

    cognized_by_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def cognized_by(cls):
        return orm.relationship(
            User,
            primaryjoin=lambda: User.uuid == cls.cognized_by_uuid,
            foreign_keys=lambda: [cls.cognized_by_uuid],
            doc="""
            Reference to the :class:`User` who last cognized the batch data.
            """)

    rowcount = sa.Column(sa.Integer(), nullable=True, doc="""
    Cached row count for the batch.  No guarantees perhaps, but should be accurate.
    """)

    complete = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag to indicate whether the batch is complete.  This may be used to assist
    with workflow when entering/executing new batches.
    """)

    executed = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date and time when the batch was (last) executed.
    """)

    executed_by_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def executed_by(cls):
        return orm.relationship(
            User,
            primaryjoin=lambda: User.uuid == cls.executed_by_uuid,
            foreign_keys=lambda: [cls.executed_by_uuid],
            doc="""
            Reference to the :class:`User` who (last) executed the batch.
            """)

    purge = sa.Column(sa.Date(), nullable=True, doc="""
    Date after which the batch may be purged.
    """)

    notes = sa.Column(sa.Text(), nullable=True, doc="""
    Any arbitrary notes for the batch.
    """)

    params = sa.Column(JSONTextDict(), nullable=True, doc="""
    Extra parameters for the batch, encoded as JSON.  This hopefully can be
    useful for "many" batches, to avoid needing to add extra flags to the
    schema etc.
    """)

    extra_data = sa.Column(sa.Text(), nullable=True, doc="""
    Extra field for arbitrary data, useful to the batch handler etc.  Note that
    there is *no* structure assumed here, it can be JSON or whatever is needed.
    """)

    STATUS_OK                   = 1
    STATUS_QUESTIONABLE         = 2

    STATUS = {
        STATUS_OK:              "ok",
        STATUS_QUESTIONABLE:    "questionable",
    }

    status_code = sa.Column(sa.Integer(), nullable=True, doc="""
    Status code for the batch as a whole.  This indicates whether the batch is
    "okay" and ready to execute, or (why) not etc.
    """)

    status_text = sa.Column(sa.String(length=255), nullable=True, doc="""
    Text which may briefly explain the batch status code, if needed.
    """)

    def __repr__(self):
        return "{}(uuid={})".format(
            self.__class__.__name__,
            repr(self.uuid))

    @property
    def id_str(self):
        if self.id:
            return "{:08d}".format(self.id)
        return ''

    def __str__(self):
        return "{} {}".format(
            self.__class__.__name__,
            self.id_str if self.id else "(new)")

    # TODO: deprecate/remove this?
    def add_row(self, row):
        """
        Convenience method for appending a data row, with auto-sequence.
        """
        self.data_rows.append(row)

    def active_rows(self):
        return [row for row in self.data_rows if not row.removed]

    def relative_filedir(self, config):
        """
        Returns the path for batch data file storage, relative to the root
        folder for all batch storage.  This includes the batch key as the first
        segment, e.g. ``'labels/d4/83d6f2aeb011e6afeb3ca9f40bc550'``.
        """
        if not self.uuid:
            orm.object_session(self).flush()
        return os.path.join(self.batch_key, self.uuid[:2], self.uuid[2:])

    # TODO: deprecate / remove this (?)
    relative_filepath = relative_filedir

    def filedir(self, config):
        """
        Returns the absolute path to the folder in which the data file resides.
        The config object determines the root path for such files, e.g.:

        .. code-block:: ini

           [rattail]
           batch.files = /path/to/batch/files

        Within this root path, a more complete path is generated using the
        :attr:`BatchMixin.key` and the :attr:`BatchMixin.uuid` values.
        """
        batchdir = config.batch_filedir()
        return os.path.abspath(os.path.join(batchdir, self.relative_filedir(config)))

    def absolute_filepath(self, config, filename=None, name_attr='filename'):
        """
        Return the absolute path where the data file resides.  This is the path
        returned by :meth:`filedir()` with the batch filename joined to it.
        """
        if filename is None:
            filename = getattr(self, name_attr)
        return os.path.join(self.filedir(config), filename)

    # for convenience
    filepath = absolute_filepath

    def filesize(self, config, name_attr='filename'):
        """
        Returns the size of the data file in bytes.
        """
        path = self.filepath(config, name_attr=name_attr)
        return os.path.getsize(path)

    def delete_data(self, config):
        """
        Delete the data folder for the batch
        """
        # TODO: should this logic be in the handler instead?
        path = config.batch_filepath(self.batch_key, self.uuid)
        if os.path.exists(path):
            shutil.rmtree(path)


class BaseFileBatchMixin(BatchMixin):
    """
    Common mixin for all batches which may involve a data file, either as
    initial data, or perhaps as "export" etc.
    """
    filename_nullable = True

    @declared_attr
    def filename(cls):
        return filename_column(nullable=cls.filename_nullable, doc="""
        Base name of the data file.
        """)


class FileBatchMixin(BaseFileBatchMixin):
    """
    Mixin for all (new-style) batch classes which involve a file upload as
    their first step.
    """
    filename_nullable = False

    def write_file(self, config, contents):
        """
        Save a data file for the batch to the location specified by
        :meth:`filepath()`.
        """
        filedir = self.filedir(config)
        if not os.path.exists(filedir):
            os.makedirs(filedir)
        with open(os.path.join(filedir, self.filename), 'wb') as f:
            f.write(contents)


class BatchRowMixin(object):
    """
    Mixin for all (new-style) batch row classes.
    """

    uuid = uuid_column()

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__()

    @classmethod
    def __batchrow_table_args__(cls):
        batch_table = cls.__batch_class__.__tablename__
        row_table = cls.__tablename__
        return (
            sa.ForeignKeyConstraint(['batch_uuid'], ['{0}.uuid'.format(batch_table)],
                                    name='{0}_fk_batch_uuid'.format(row_table)),
        )

    @classmethod
    def __default_table_args__(cls):
        return cls.__batchrow_table_args__()

    STATUS = {}

    batch_uuid = sa.Column(sa.String(length=32), nullable=False)

    @declared_attr
    def batch(cls):
        batch_class = cls.__batch_class__
        row_class = cls
        batch_class.row_class = row_class

        # Must establish `Batch.data_rows` here instead of from within `Batch`
        # itself, because the row class doesn't yet exist when that happens.
        batch_class.data_rows = orm.relationship(
            row_class,
            order_by=lambda: row_class.sequence,
            collection_class=ordering_list('sequence', count_from=1),
            cascade='all, delete-orphan',
            doc="""
            Collection of data rows for the batch.

            .. note::
               I would prefer for this attribute to simply be named "rows"
               instead of "data_rows", but unfortunately (as of this writing)
               "rows" is essentially a reserved word in FormAlchemy.
            """,
            back_populates='batch')

        # Now, here's the `BatchRow.batch` reference.
        return relationship(batch_class, back_populates='data_rows', doc="""
        Reference to the parent batch to which the row belongs.
        """)

    sequence = sa.Column(sa.Integer(), nullable=False, doc="""
    Sequence number of the row within the batch.  This number should be from 1 to
    the actual number of rows in the batch.
    """)

    status_code = sa.Column(sa.Integer(), nullable=True, doc="""
    Status code for the data row.  This indicates whether the row's product could
    be found in the system, etc.  Ultimately the meaning of this is defined by each
    particular batch type.
    """)

    status_text = sa.Column(sa.String(length=255), nullable=True, doc="""
    Short description of row status.  Ultimately the meaning and use of this is
    defined by each particular batch type.
    """)

    modified = sa.Column(sa.DateTime(), nullable=True, default=make_utc, onupdate=make_utc, doc="""
    Last modification time of the row.  This should be automatically set when
    the row is first created, as well as anytime it's updated thereafter.
    """)

    removed = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag to indicate a row has been removed from the batch.
    """)


class ProductBatchRowMixin(BatchRowMixin):
    """
    Mixin for all row classes of (new-style) batches which pertain to products.
    """

    @classmethod
    def __default_table_args__(cls):
        batch_table = cls.__batch_class__.__tablename__
        row_table = cls.__tablename__
        return (
            sa.ForeignKeyConstraint(['batch_uuid'], ['{0}.uuid'.format(batch_table)],
                                    name='{0}_fk_batch'.format(row_table)),
            sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'],
                                    name='{0}_fk_product'.format(row_table)),
            )

    item_entry = sa.Column(sa.String(length=20), nullable=True, doc="""
    Raw entry value, as obtained from the initial data source, and which is
    used to locate the product within the system.  This raw value is preserved
    in case the initial lookup fails and a refresh must attempt further
    lookup(s) later.  Only used by certain batch handlers in practice.
    """)

    upc = sa.Column(GPCType(), nullable=True, doc="""
    UPC for the product associated with the row.
    """)

    item_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    Generic ID string for the product associated with the row.
    """)

    product_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def product(self):
        table_name = self.__batch_class__.__tablename__
        model_title = self.__batch_class__.get_model_title()
        return orm.relationship(
            Product,
            doc="""
            Reference to the product with which the row is associated, if any.
            """,
            backref=orm.backref(
                '_{}_rows'.format(table_name),
                doc="""
                Sequence of all {} rows which reference the product.
                """.format(model_title)))

    brand_name = sa.Column(sa.String(length=100), nullable=True, doc="""
    Brand name of the product.
    """)

    description = sa.Column(sa.String(length=255), nullable=True, doc="""
    Description of the product.
    """)

    size = sa.Column(sa.String(length=255), nullable=True, doc="""
    Size of the product, as string.
    """)

    # TODO: should add this probably, but for now one batch has Integer for this..
    # case_quantity = sa.Column(sa.Numeric(precision=6, scale=2), nullable=True, doc="""
    # Number of units in a case of product.
    # """)

    department_number = sa.Column(sa.Integer(), nullable=True, doc="""
    Number of the department to which the product belongs.
    """)

    department_name = sa.Column(sa.String(length=30), nullable=True, doc="""
    Name of the department to which the product belongs.
    """)

    subdepartment_number = sa.Column(sa.Integer(), nullable=True, doc="""
    Number of the subdepartment to which the product belongs.
    """)

    subdepartment_name = sa.Column(sa.String(length=30), nullable=True, doc="""
    Name of the subdepartment to which the product belongs.
    """)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Label Printing
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy.orm import object_session

from .core import Base, uuid_column
from rattail.util import load_object


@six.python_2_unicode_compatible
class LabelProfile(Base):
    """
    Represents a collection of settings for product label printing.
    """
    __tablename__ = 'label_profile'
    __versioned__ = {}

    uuid = uuid_column()
    ordinal = sa.Column(sa.Integer())

    code = sa.Column(sa.String(length=30), nullable=True, doc="""
    Supposedly unique "code" for the label profile.  May be useful for
    identification of a common label type across nodes, for instance.
    """)

    description = sa.Column(sa.String(length=50))
    printer_spec = sa.Column(sa.String(length=255))
    formatter_spec = sa.Column(sa.String(length=255))
    format = sa.Column(sa.Text())
    visible = sa.Column(sa.Boolean())

    sync_me = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether this label profile should be synced across "all"
    other Rattail systems across the organization.
    """)

    _printer = None
    _formatter = None

    def __str__(self):
        return str(self.description or '')

    # TODO: Copy elsewhere, deprecate and remove this logic (not strictly an
    # API call from within data model, but should go with the others, below).
    def get_formatter(self, config):
        if not self._formatter and self.formatter_spec:
            try:
                formatter = load_object(self.formatter_spec)
            except AttributeError:
                pass
            else:
                self._formatter = formatter(config)
                self._formatter.format = self.format
        return self._formatter        

    # TODO: Copy elsewhere, deprecate and remove this logic (API call from
    # within data model).
    def get_printer(self, config):
        if not self._printer and self.printer_spec:
            try:
                printer = load_object(self.printer_spec)
            except AttributeError:
                pass
            else:
                self._printer = printer(config)
                for name in printer.required_settings:
                    setattr(printer, name, self.get_printer_setting(name))
                self._printer.formatter = self.get_formatter(config)
        return self._printer

    # TODO: Copy elsewhere, deprecate and remove this logic (API call from
    # within data model).
    def get_printer_setting(self, name):
        from rattail.db.api import get_setting
        if self.uuid is None:
            return None
        session = object_session(self)
        name = 'labels.{0}.printer.{1}'.format(self.uuid, name)
        return get_setting(session, name)

    # TODO: Copy elsewhere, deprecate and remove this logic (API call from
    # within data model).
    def save_printer_setting(self, name, value):
        from rattail.db.api import save_setting
        session = object_session(self)
        if self.uuid is None:
            session.flush()
        name = 'labels.{0}.printer.{1}'.format(self.uuid, name)
        save_setting(session, name, value)



# -*- coding: utf-8 -*-
"""add receiving.truck_dump_batch

Revision ID: 6551a4d9ff25
Revises: c37154504ae1
Create Date: 2018-05-16 10:40:14.855471

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '6551a4d9ff25'
down_revision = u'c37154504ae1'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('truck_dump_batch_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'purchase_batch_fk_truck_dump_batch', 'purchase_batch', 'purchase_batch', ['truck_dump_batch_uuid'], ['uuid'], use_alter=True)


def downgrade():

    # purchase_batch
    op.drop_constraint(u'purchase_batch_fk_truck_dump_batch', 'purchase_batch', type_='foreignkey')
    op.drop_column('purchase_batch', 'truck_dump_batch_uuid')
